import torch
from sklearn.base import TransformerMixin
from sklearn.impute import SimpleImputer as _SimpleImputer
from torchcde import linear_interpolation_coeffs

from ._mixin import apply_fit_to_channels, apply_transform_to_channels, apply_transform_to_channel_subset


class SimpleImputer(TransformerMixin):
    """Basic imputation for tensors. Simply borrows from sklearns SimpleImputer.

    Assumes the size is (..., length, input_channels), reshapes to (..., input_channels), performs the method
    operation and then reshapes back.

    Arguments:
        strategy (str): One of ('mean', 'median', 'most_frequent', 'constant').
        fill_value (float): The value to fill nans with, this is active only if `strategy = 'constant'`.
    """

    def __init__(self, strategy, fill_value, channel_indices=None):
        self.strategy = strategy
        self.fill_value = fill_value
        self.channel_indices = channel_indices

        self.imputer = _SimpleImputer(strategy=strategy, fill_value=fill_value)

    @apply_fit_to_channels
    def fit(self, data, labels=None):
        self.imputer.fit(data)
        return self

    @apply_transform_to_channel_subset
    @apply_transform_to_channels
    def transform(self, data):
        output_data = torch.Tensor(self.imputer.transform(data))
        return output_data


class NegativeFilter(TransformerMixin):
    """Replace negative values with zero.

    Arguments:
        fill_value (float): The values to replace the negative values with.
    """

    def __init__(self, fill_value=float("nan"), channel_indices=None):
        self.fill_value = fill_value
        self.channel_indices = channel_indices

    def fit(self, data, labels=None):
        return self

    @apply_transform_to_channel_subset
    def transform(self, data):
        data[data < 0] = self.fill_value
        return data


class Interpolation(TransformerMixin):
    """Perform linear (or rectilinear) interpolation on the missing values.

    Arguments:
        method (str): One of 'linear' or 'rectilinear'.
    """

    def __init__(self, method="linear", channel_indices=None):
        assert method in ["linear", "rectilinear"], "Got method {} which is not recognised".format(method)
        self.method = method
        self.channel_indices = channel_indices

        # Linear interpolation function requires the channel index of times
        self._rectilinear = 0 if self.method == "rectilinear" else None

    def fit(self, data, labels=None):
        return self

    @apply_transform_to_channel_subset
    def transform(self, data):
        return linear_interpolation_coeffs(data, rectilinear=self._rectilinear)


class ForwardFill(TransformerMixin):
    """Forward fills data in a torch tensor of shape (..., length, input_channels) along the length dim.

    Arguments:
        fill_index (int): Denotes the index to fill down. Default is -2 as we tend to use the convention (..., length,
            input_channels) filling down the length dimension.
        backwards (bool): Set True to first flip the tensor along the length axis so as to perform a backwards fill.
        channel_indices (list or None): Leave as None to apply to the whole dataset, else apply a list of indices to
            apply only to that subset of channels, e.g. [0, 2, 5] will apply to the first third and fifth channel.

    Example:
        >>> x = torch.tensor([[1, 2], [float('nan'), 1], [2, float('nan')]], dtype=torch.float)
        >>> ForwardFill().transform(x, fill_index=-2, backwards=False)
        tensor([
            [1., 2.],
            [1., 1.],
            [2., 1.]
        ])

    Returns:
        A tensor with forward filled data.
    """

    def __init__(self, fill_index=-2, backwards=False, channel_indices=None):
        self.fill_index = fill_index
        self.backwards = backwards
        self.channel_indices = channel_indices

    def fit(self, data, labels=None):
        return self

    @apply_transform_to_channel_subset
    def transform(self, data):
        return _forward_fill(data, self.fill_index, self.backwards)


def _forward_fill(x, fill_index=-2, backwards=False):
    # Checks
    assert isinstance(x, torch.Tensor)
    assert x.dim() >= 2

    # flipping if a backwards fill
    def backflip(x):
        x = x.flip(fill_index) if backwards else x
        return x

    x = backflip(x)

    mask = torch.isnan(x)
    if mask.any():
        cumsum_mask = (~mask).cumsum(dim=fill_index)
        cumsum_mask[mask] = 0
        _, index = cumsum_mask.cummax(dim=fill_index)
        x = x.gather(dim=fill_index, index=index)

    # Re-flip if backwards
    x = backflip(x)

    return x



import torch

from tsbatteries.preprocessing import impute


def test_forward_fill():
    """ Test forward fill against the results returned from a nested for loop. """
    devices = ["cpu"]
    if torch.cuda.is_available():
        devices.append("cuda")

    for device in devices:
        # Check ffill
        for N, L, C in [(1, 5, 3), (2, 2, 2), (3, 2, 1)]:
            x = torch.randn(N, L, C).to(device)
            # Drop mask
            tensor_num = x.numel()
            mask = torch.randperm(tensor_num)[: int(0.3 * tensor_num)].to(device)
            x.view(-1)[mask] = float("nan")
            x_ffilled = x.clone().float()
            for i in range(0, x.size(0)):
                for j in range(x.size(1)):
                    for k in range(x.size(2)):
                        non_nan = x_ffilled[i, : j + 1, k][
                            ~torch.isnan(x[i, : j + 1, k])
                        ]
                        input_val = (
                            non_nan[-1].item() if len(non_nan) > 0 else float("nan")
                        )
                        x_ffilled[i, j, k] = input_val
            x_ffilled_actual = impute._forward_fill(x)
            assert x_ffilled.allclose(x_ffilled_actual, equal_nan=True)



import os
import sys

sys.path.insert(0, os.path.abspath("../"))


# Info
project = "Batteries"
copyright = "2021, James Morrill"
author = "James Morrill"
release = "v0.0.1"

# Sphinx extensions
extensions = ["sphinx.ext.autodoc", "sphinx.ext.coverage", "sphinx.ext.napoleon"]

# Things
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
templates_path = ["_templates"]

# Dont include the full modules path
add_module_names = False

# Dont do alphabetical
autodoc_member_order = "bysource"

# Use an edited version of the catalyst theme
# Answer taken from
# https://stackoverflow.com/questions/14622698/customize-sphinxdoc-theme
html_theme = "theme"  # use the theme in subdir 'theme'
html_theme_path = ["."]  # make sphinx search for themes in current dir

# Themes
# html_theme = "karma_sphinx_theme"
# html_theme = 'catalyst_sphinx_theme'
# html_theme = 'pytorch_sphinx_theme'
# html_theme = 'sphinx_rtd_theme'



def apply_transform_to_channels(transform):
    """Decorator for transforms that need to be applied along the channels only.

    Assumes the size is (..., length, input_channels), reshapes to (..., input_channels), performs the method
    operation and then reshapes back.

    For example, suppose we wish to use the StandardScaler on a 3d tensor, we can wrap the transform with this
    decorator and it will first reshape to (-1, input_channels), apply the operation, then reshape back.

    Arguments:
        transform (method): An sklearn-type transform method.

    Returns:
        The transformed data in the original shape.
    """

    def inner(self, data):
        # Reshape to 2d
        data_2d = data.reshape(-1, data.size(-1))

        # Apply the transform
        transformed_data = transform(self, data_2d)

        # Reshape back
        output = transformed_data.reshape(data.size())

        return output

    return inner


def apply_fit_to_channels(fit):
    """ See `apply_transform_to_channels`, this function is for the fit method. """

    def inner(self, data, labels=None):
        # Reshape to 2d and return the fit
        data_2d = data.reshape(-1, data.size(-1))
        return fit(self, data_2d, labels=labels)

    return inner


def apply_transform_to_channel_subset(transform):
    """Decorator for applying to only a subset of the channel indices.

    This requires the class to have a channel_indices argument.
    """

    def inner(self, data):
        assert hasattr(self, 'channel_indices'), "Decorator requires channel_indices attribute."
        if self.channel_indices is None:
            data = transform(self, data)
        else:
            data[..., self.channel_indices] = transform(self, data[..., self.channel_indices])
        return data

    return inner



"""
problem.py
===============================
Methods for defining the problem.
"""
import numpy as np


def _assertions(data, label_index, data_columns, label_column, problem_type):
    # Assertions for `generate_labels`
    if problem_type == "online":
        assert data.dim() == 3
    elif problem_type == "oneshot":
        assert data.dim() == 2
    else:
        raise NotImplementedError(
            "Problem type must be 'online' or 'oneshot', recieved {}".format(
                problem_type
            )
        )

    # Ensure only one of label_index or label column is specified
    assert (label_index is None) ^ (label_column is None), (
        "Only one of `label_index` or `label_column` can be " "specified."
    )
    if label_column is not None:
        assert (
            label_column in data_columns
        ), "Cannot find `label_column` within `data_columns."
        assert (
            data_columns is not None
        ), "`data_columns` must be specified with `label_column`."
        col_size, data_size = len(data_columns), data.size(-1)
        assert col_size == data_size, (
            "`data_columns` must be the same length as the channel dim. Mismatched sizes "
            "{} and {}.".format(col_size, data_size)
        )


def _create_online_labels(labels, lookback=0, lookforward=0):
    """ Loop over the ids and fill in labels according to the parameters. """
    for i in range(len(labels)):
        labels_ = labels[i]
        for j in np.argwhere(labels_ == i):
            labels_[j - lookback : j + lookforward] = 1
        labels[i] = labels_
    return labels


def generate_labels(
    data,
    label_index=None,
    data_columns=None,
    label_column=None,
    problem_type=None,
):
    # Sanity
    _assertions(data, label_index, data_columns, label_column, problem_type)

    # Get the relevant labels and remove them from the data
    if label_column is not None:
        label_index = [
            i for i, column in enumerate(data_columns) if column == label_column
        ][0]
    data_indexes = [x for x in range(data.size(-1)) if x != label_index]
    labels = data[..., label_index]
    data = data[..., data_indexes]

    # Now reformat the labels in the case where the problem is online
    if problem_type == "online":
        labels = _create_online_labels(labels)

    return data, labels



import torch
from sklearn.base import TransformerMixin
from sklearn.preprocessing import (FunctionTransformer, MaxAbsScaler,
                                   MinMaxScaler, StandardScaler)

from ._mixin import apply_fit_to_channels, apply_transform_to_channels

SCALERS = {
    "stdsc": StandardScaler,
    "ma": MaxAbsScaler,
    "mms": MinMaxScaler,
}


class TensorScaler(TransformerMixin):
    """Scaling for 3D tensors.

    Assumes the size is (..., length, input_channels), reshapes to (..., input_channels), performs the method
    operation and then reshapes back.

    Arguments:
        method (str): Scaling method, one of ('stdsc', 'ma', 'mms').
        scaling_function (transformer): Specification of an sklearn transformer that performs a scaling operation.
            Only one of this or scaling can be specified.
    """

    def __init__(self, method="stdsc", scaling_function=None):
        self.scaling = method

        if all([method is None, scaling_function is None]):
            self.scaler = FunctionTransformer(func=None)
        elif isinstance(method, str):
            self.scaler = SCALERS.get(method)()
            assert (
                self.scaler is not None
            ), "Scalings allowed are {}, recieved {}.".format(SCALERS.keys(), method)
        else:
            self.scaler = scaling_function

    @apply_fit_to_channels
    def fit(self, data, labels=None):
        self.scaler.fit(data)
        return self

    @apply_transform_to_channels
    def transform(self, data):
        output_data = torch.Tensor(self.scaler.transform(data))
        return output_data


def scale_tensors(tensors, method="stdsc"):
    """A function version of TensorScaler, if multiple tensors are specified the first is used to fit the scaler

    Arguments:
        tensors (tensor or list of tensors): Data to transform. If multiple tensors are specified, the first is used to
            fit.
        method (str): See TensorScaler.

    Returns:
        Same format as input only now scaled.
    """
    scaler = TensorScaler(method=method)

    if isinstance(tensors, list):
        scaler.fit(tensors[0])
        output = [scaler.transform(x) for x in tensors]
    else:
        output = scaler.fit_transform(tensors)

    return output



import torch
from torch import nn


class RETAIN(nn.Module):
    """RETAIN model with time-series capabilities. """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        output_dim,
        embedding_kwargs=None,
        return_sequences=False,
    ):
        super(RETAIN, self).__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.embedding_kwargs = embedding_kwargs
        self.return_sequences = return_sequences

        if self.return_sequences:
            raise NotImplementedError

        # Embedding dim if specified
        if embedding_kwargs is not None:
            self.setup_embedding_layer()

        # For extraction of most relevant series information
        self.series_sequential = nn.Sequential(
            nn.GRU(input_dim, hidden_dim, batch_first=True),
            _ItemSelector(0),
            nn.Linear(hidden_dim, 1),
            nn.Softmax(),
        )

        # For extraction of most relevant features
        self.features_sequential = nn.Sequential(
            nn.GRU(input_dim, hidden_dim, batch_first=True),
            _ItemSelector(0),
            nn.Linear(hidden_dim, input_dim),
            nn.Tanh(),
        )

        # Output layer
        self.fc_output = nn.Linear(input_dim, output_dim)

    def _setup_embedding(self, embedding_kwargs):
        """ Setup an embedding if args are specified. """
        kwarg_keys = ["input_dropout", "embedding_dim", "embedding_dropout"]
        if any([embedding_kwargs.get(x) is None for x in kwarg_keys]):
            raise AssertionError(
                "Invalid keys for embedding_kwargs, must contain {}".format(kwarg_keys)
            )
        self.embedding = nn.Sequential(
            nn.Dropout(embedding_kwargs["input_dropout"]),
            nn.Linear(self.input_dim, embedding_kwargs["embedding_dim"]),
            nn.Dropout(embedding_kwargs["embedding_dropout"]),
        )
        self.input_dim = embedding_kwargs["embedding_dim"]

    @staticmethod
    def _generate_context(x, alpha, beta):
        """ Method for generating the context vector. """
        return torch.bmm(torch.transpose(alpha, 1, 2), beta * x).squeeze(1)

    def forward(self, x):
        if self.embedding_kwargs is not None:
            x = self.embedding(x)

        # Get series info
        alpha = self.series_sequential(x)

        # Get feature info
        beta = self.features_sequential(x)

        # Apply the final sum(alpha \times (beta dot x))
        context = self._generate_context(x, alpha, beta)

        # Linear layer and output
        output = self.fc_output(context)

        return output


class _ItemSelector(nn.Module):
    """ Used for extracting an item from a nn.Module that returns a tuple, such as an RNN. """

    def __init__(self, index):
        super(_ItemSelector, self).__init__()
        self.index = index

    def forward(self, x):
        return x[self.index]



from sklearn.pipeline import Pipeline

from .impute import ForwardFill, Interpolation, NegativeFilter, SimpleImputer
from .misc import PadRaggedTensors
from .pipeline import PipelineCompiler, PipelineDataset
from .scale import TensorScaler
from .split import tensor_train_test_split, train_val_test_split

__all__ = [
    # pipe
    "Pipeline",
    "PipelineCompiler",
    "PipelineDataset",
    # misc
    "PadRaggedTensors",
    # impute
    "NegativeFilter",
    "Interpolation",
    "ForwardFill",
    "SimpleImputer",
    # Scale
    "TensorScaler",
    # Split
    "tensor_train_test_split",
    "train_val_test_split",
]



import torch


def make_classification_problem(
    num_samples=50, length=10, input_dim=3, test_frac=0.4, static_dim=None
):
    # Random data, binary classification problem
    data = torch.randn((num_samples, length, input_dim))
    labels = torch.randint(0, 2, (num_samples, 1))

    # No negative values for labels == 1, so the task can train effectively
    positive_data = data[(labels == 1).view(-1)]
    positive_data[positive_data < 0] = 0
    data[(labels == 1).view(-1)] = positive_data

    # Split train test
    test_start = int(((1 - test_frac) * num_samples))
    train_data = [data[:test_start], labels[:test_start]]
    test_data = [data[test_start:], labels[test_start:]]

    # Add static if set
    if static_dim is not None:
        static_data = torch.randn(num_samples, static_dim)
        train_data = [static_data[:test_start]] + train_data
        test_data = [static_data[test_start:]] + test_data

    return train_data, test_data



import pytest

from tsbatteries.models import NeuralCDE

from .helpers import make_classification_problem
from .test_models import training_loop


def create_ncde_problem(static_dim=None, use_initial=True):
    # Simple problem
    train_data, test_data = make_classification_problem(static_dim=static_dim)

    # Get sizes
    input_dim, output_dim = [t.size(-1) for t in train_data][-2:]

    # Setup and NCDE
    hidden_dim = 15
    model = NeuralCDE(
        input_dim,
        hidden_dim,
        output_dim,
        static_dim=static_dim,
        interpolation="linear",
        use_initial=use_initial,
    )

    # Function to convert to ncde format data
    def to_ncde(tensors):
        if static_dim is not None:
            tensors = (tensors[0], tensors[1]), tensors[2]
        return tensors

    return model, to_ncde(train_data), to_ncde(test_data)


@pytest.mark.parametrize(
    "static_dim, use_initial",
    [(None, True), (None, False), (5, True), (5, False)],
)
def test_ncde_simple(static_dim, use_initial):
    # Test the model runs and gets a normal accuracy
    model, train_data, _ = create_ncde_problem(
        static_dim=static_dim, use_initial=use_initial
    )
    _, acc = training_loop(model, *train_data, n_epochs=10)
    assert 0 <= acc.item() <= 1



import torch
import torchcde
from torch import nn

SPLINES = {
    "cubic": torchcde.NaturalCubicSpline,
    "linear": torchcde.LinearInterpolation,
    "rectilinear": torchcde.LinearInterpolation,
}


class NeuralCDE(nn.Module):
    """Performs the Neural CDE training process over a batch of time series.

    Arguments:
        input_dim (int): The dimension of the path.
        hidden_dim (int): The dimension of the hidden state.
        output_dim (int): The dimension of the output.
        static_dim (int): The dimension of any static values, these will be concatenated to the initial values and
            put through a network to build h0.
        hidden_hidden_dim (int): The dimension of the hidden layer in the RNN-like block.
        num_layers (int): The number of hidden layers in the vector field. Set to 0 for a linear vector field.
            net with the given density. Hidden and hidden hidden dims must be multiples of 32.
        use_initial (bool): Set True to use the initial absolute values to generate h0.
        adjoint (bool): Set True to use odeint_adjoint.
        solver (str): ODE solver, must be implemented in torchdiffeq.
        return_sequences (bool): If True will return the linear function on the final layer, else linear function on
            all layers.
        apply_final_linear (bool): Set False for no final linear layer to be applied to the hidden state.
        return_filtered_rectilinear (bool): Set True to return every other output if the interpolation scheme chosen
            is rectilinear, this is because rectilinear doubles the input length. False will return the full output.
    """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        output_dim,
        static_dim=None,
        hidden_hidden_dim=15,
        num_layers=3,
        use_initial=True,
        interpolation="linear",
        adjoint=True,
        solver="rk4",
        return_sequences=False,
        apply_final_linear=True,
        return_filtered_rectilinear=True,
    ):
        super().__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.static_dim = static_dim
        self.hidden_hidden_dim = hidden_hidden_dim
        self.num_layers = num_layers
        self.use_initial = use_initial
        self.interpolation = interpolation
        self.adjoint = adjoint
        self.solver = solver
        self.return_sequences = return_sequences
        self.apply_final_linear = apply_final_linear
        self.return_filtered_rectilinear = return_filtered_rectilinear

        # Set initial linear layer
        if self.initial_dim > 0:
            self.initial_linear = nn.Linear(self.initial_dim, self.hidden_dim)

        # Interpolation function
        assert (
            self.interpolation in SPLINES.keys()
        ), "Unrecognised interpolation scheme {}".format(self.interpolation)
        self.spline = SPLINES.get(self.interpolation)

        # The net that is applied to h_{t-1}
        self.func = _NCDEFunc(input_dim, hidden_dim, hidden_hidden_dim, num_layers)

        # Linear classifier to apply to final layer
        self.final_linear = (
            nn.Linear(self.hidden_dim, self.output_dim)
            if apply_final_linear
            else lambda x: x
        )

    @property
    def initial_dim(self):
        # Setup initial dim dependent on `use_initial` and `static_dim` options
        initial_dim = 0
        if self.use_initial:
            initial_dim += self.input_dim
        if self.static_dim is not None:
            initial_dim += self.static_dim
        return initial_dim

    def _setup_h0(self, inputs):
        """Sets up the initial value of the hidden state.

        The hidden state depends on the options `use_initial` and `static_dim`. If either of these are specified the
        hidden state will be generated via a network applied to either a concatenation of the initial and static data,
        or a network applied to just initial/static depending on options. If neither are specified then a zero initial
        hidden state is used.
        """
        if self.static_dim is None:
            coeffs = inputs
            if self.use_initial:
                h0 = self.initial_linear(inputs[..., 0, :])
            else:
                h0 = torch.autograd.Variable(
                    torch.zeros(inputs.size(0), self.hidden_dim)
                ).to(inputs.device)
        else:
            assert (
                len(inputs) == 2
            ), "Inputs must be a 2-tuple of (static_data, temporal_data)"
            static, coeffs = inputs
            if self.use_initial:
                h0 = self.initial_linear(
                    torch.cat((static, coeffs[..., 0, :]), dim=-1)
                )
            else:
                h0 = self.initial_linear(static)

        return coeffs, h0

    def _make_outputs(self, hidden):
        """ Hidden state to output format depending on `return_sequences` and rectilinear (return every other). """
        if self.return_sequences:
            outputs = self.final_linear(hidden)

            # If rectilinear and return sequences, return every other value
            if (self.interpolation == "rectilinear") and self.return_filtered_rectilinear:
                outputs = outputs[:, ::2]
        else:
            outputs = self.final_linear(hidden[:, -1, :])
        return outputs

    def forward(self, inputs):
        # Handle h0 and inputs
        coeffs, h0 = self._setup_h0(inputs)

        # Make lin int
        data = self.spline(coeffs)

        # Perform the adjoint operation
        hidden = torchcde.cdeint(
            data,
            self.func,
            h0,
            data.grid_points,
            adjoint=self.adjoint,
            method=self.solver,
        )

        # Convert to outputs
        outputs = self._make_outputs(hidden)

        return outputs


class _NCDEFunc(nn.Module):
    """The function applied to the hidden state in the NCDE model.

    This creates a simple RNN-like block to be used as the computation function f in:
        dh/dt = f(h) dX/dt
    """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        hidden_hidden_dim=15,
        num_layers=1,
        density=0.0,
        rank=None,
    ):
        super().__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.hidden_hidden_dim = hidden_hidden_dim
        self.num_layers = num_layers
        self.sparsity = density
        self.rank = rank

        # Additional layers are just hidden to hidden with relu activation
        layers = [nn.Linear(hidden_dim, hidden_hidden_dim), nn.ReLU()]
        if num_layers > 1:
            for _ in range(num_layers - 1):
                layers += [nn.Linear(hidden_hidden_dim, hidden_hidden_dim), nn.ReLU()]

        # Add on final layer and Tanh and build net
        layers += [nn.Linear(hidden_hidden_dim, hidden_dim * input_dim), nn.Tanh()]
        self.net = nn.Sequential(*layers)

    def forward(self, t, h):
        return self.net(h).view(-1, self.hidden_dim, self.input_dim)



from torch import nn

MODELS = {"rnn": nn.RNN, "gru": nn.GRU, "lstm": nn.LSTM}


class RNN(nn.Module):
    """Standard RNN.

    Arguments:
        input_dim (int): The dimension of the path.
        hidden_dim (int): The dimension of the hidden state.
        output_dim (int): The dimension of the output.
        num_layers (int): The number of hidden layers in the vector field. Set to 0 for a linear vector field.
            net with the given density. Hidden and hidden hidden dims must be multiples of 32.
        model_string (int): Any of ('rnn', 'gru', 'lstm')
        nonlinearity (str): One of ('tanh', 'relu').
        bias (bool): Whether to add a bias term.
        return_sequences (bool): If True will return the linear function on the final layer, else linear function on
            all layers.
        apply_final_linear (bool): Set False for no final linear layer to be applied to the hidden state.
    """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        output_dim,
        num_layers=1,
        model_string="rnn",
        nonlinearity="tanh",
        bias=True,
        dropout=0,
        return_sequences=True,
        apply_final_linear=True,
    ):
        super(RNN, self).__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.num_layers = num_layers
        self.model_string = model_string
        self.nonlinearity = nonlinearity
        self.bias = bias
        self.dropout = dropout
        self.return_sequences = return_sequences
        self.apply_final_linear = apply_final_linear

        model = MODELS.get(model_string)
        if model is None:
            raise NotImplementedError(
                "model_string must be one of {}, got {}".format(
                    MODELS.keys(), model_string
                )
            )
        self.rnn = model(
            input_size=input_dim,
            hidden_size=hidden_dim,
            num_layers=num_layers,
            nonlinearity=nonlinearity,
            bias=bias,
            dropout=dropout,
            batch_first=True,
        )
        self.total_hidden_size = num_layers * hidden_dim
        self.final_linear = (
            nn.Linear(self.total_hidden_size, output_dim)
            if self.apply_final_linear
            else lambda x: x
        )

    def forward(self, x):
        # Run the RNN
        h_full, _ = self.rnn(x)

        # Terminal output if classifcation else return all outputs
        outputs = (
            self.final_linear(h_full[:, -1, :])
            if not self.return_sequences
            else self.final_linear(h_full)
        )

        return outputs



import numpy as np
import pytest
import torch

from tsbatteries import preprocessing
from tsbatteries.preprocessing.pipeline import (PipelineCompiler,
                                                PipelineDataset)
from tsbatteries.tests.helpers import make_classification_problem


@pytest.fixture()
def static_pipeline():
    pipeline = preprocessing.Pipeline(
        [
            ("negative_impute", preprocessing.NegativeFilter()),
            ("stdsc", preprocessing.TensorScaler(method="mms")),
        ]
    )
    return pipeline


@pytest.fixture()
def temporal_pipeline():
    pipeline = preprocessing.Pipeline(
        [
            ("pad", preprocessing.PadRaggedTensors()),
            ("negative_impute", preprocessing.NegativeFilter()),
            ("stdsc", preprocessing.TensorScaler(method="stdsc")),
            ("interpolation", preprocessing.Interpolation(method="linear")),
            ("backfill", preprocessing.ForwardFill(backwards=True)),
        ]
    )
    return pipeline


def test_pipeline(temporal_pipeline):
    # Create random tensor data
    data = [torch.randn(5, 2) for _ in range(5)]
    assert torch.isnan(temporal_pipeline.fit_transform(data)).sum() == 0


@pytest.mark.parametrize(
    "static_dim, batch_size",
    [
        (None, None),
        (None, 32),
        (5, None),
        (5, 32),
    ],
)
def test_preprocessing_pipeline(
    static_pipeline, temporal_pipeline, static_dim, batch_size
):
    # Check the preprocessing pipeline can be run effectively with and without static data
    train_data, _ = make_classification_problem(static_dim=5)
    static_data, temporal_data, labels = train_data
    temporal_data = np.array([x for x in temporal_data], dtype=np.object)
    if static_dim is None:
        static_data = None
        static_pipeline = None

    # Check fit transform works with full data
    dataset = PipelineDataset(static_data, temporal_data, labels)
    main_pipeline = PipelineCompiler(
        static_pipeline, temporal_pipeline, batch_size=batch_size
    )
    output = main_pipeline.fit_transform(dataset)

    # Check fit transform works with and without static
    assert len(output) == 3



import torch

from tsbatteries.preprocessing import misc


def test_ragged_tensor_list_to_tensor():
    # This just calls pack padded sequence so we shouldnt have problems with it
    # Setup an unequal length tensor list
    list_arrays = [
        [[1, 2, 3], [3, 4, float("nan")]],
        [[1, 2, float("nan")], [5, 6, 1], [9, 10, 2]],
    ]
    tensor_list = [torch.tensor(x, dtype=float) for x in list_arrays]

    # True solution given fill
    def get_truth(fill_value):
        a1 = torch.cat([tensor_list[0], fill_value * torch.ones(1, 3)], dim=0)
        return torch.stack([a1, tensor_list[1]])

    for fill_value in [float("nan"), 0.0]:
        tensor, length = misc._pad_ragged_tensors(tensor_list, fill_value=fill_value)
        assert tensor.allclose(get_truth(fill_value), equal_nan=True)



from sklearn.base import TransformerMixin
from torch.utils.data import DataLoader, TensorDataset

from tsbatteries.models.utils import StaticTemporalTensorDataset
from tsbatteries.preprocessing.split import train_val_test_split


class PipelineDataset:
    """ Class that holds static, temporal and label data that can be fed into the PipelineCompiler. """

    def __init__(self, static_data=None, temporal_data=None, labels=None):
        """
        Args:
            static_data (tensor): Static data of shape
            temporal_data (tensor or list of tensors):
            labels (tensor):
        """
        self.static_data = static_data
        self.temporal_data = temporal_data
        self.labels = labels

    def __len__(self):
        return len(self.temporal_data)


class PipelineCompiler(TransformerMixin):
    """Combines preprocessing pipelines and data splitting.

    Given a PipelineDataset, the fit method will first split the data into train/val/test, then apply the pipelines to
    the relevant data on the training set only. The split indices are saved as attributes that will be used in the
    transform to return the transforms on each split.

    Attributes:
        train_indices (list): The index locations of the training data on the fitted dataset.
        val_indices (list): As train_indices but validation.
        test_indices (list): As train_indices but test.
        indicies (list): A list containing [train_indices, val_indices, test_indices]. Useful for easy iteration.
    """

    def __init__(self, static_pipeline=None, temporal_pipeline=None, batch_size=None):
        """
        Args:
            static_pipeline (sklearn pipeline): The pipeline to be applied to the static data. Leave as none if no
                static data is being used.
            temporal_pipeline (sklearn pipeline): The pipeline to be applied to the temporal data.
            batch_size (int): Set this only if you wish the output of the transform method to be three dataloaders
                rather than tensor lists.
        """
        self.static_pipeline = static_pipeline
        self.temporal_pipeline = temporal_pipeline
        self.batch_size = batch_size

        self.train_indices = None
        self.val_indices = None
        self.test_indices = None

    @property
    def indices(self):
        return [self.train_indices, self.val_indices, self.test_indices]

    def fit(self, dataset):
        """ Applies the train/val/test split and then fits the pipeline to the training data. """
        # Split
        self.train_indices, self.val_indices, self.test_indices = train_val_test_split(
            [dataset.labels], return_indices=True, stratify_idx=0
        )

        # Apply the pipelines to the training data
        if dataset.static_data is not None:
            self.static_pipeline.fit(dataset.static_data[self.train_indices])
        self.temporal_pipeline.fit(dataset.temporal_data[self.train_indices])

        return self

    def _apply_pipeline_to_indices(self, dataset, indices):
        # Get temporal and labels
        temporal_data = self.temporal_pipeline.transform(dataset.temporal_data[indices])
        labels = dataset.labels[indices]
        tensors = [temporal_data, labels]

        # Finally add static if we have static data
        if dataset.static_data is not None:
            static_data = self.static_pipeline.transform(dataset.static_data[indices])
            tensors = [static_data] + tensors

        return tensors

    def _to_dataloader(self, tensors):
        # Convert into a dataloader
        dataset_class = (
            StaticTemporalTensorDataset if len(tensors) == 3 else TensorDataset
        )
        dataset = dataset_class(*tensors)
        dataloader = DataLoader(dataset, batch_size=self.batch_size)
        return dataloader

    def transform(self, dataset):
        """Applies the pipeline transformation methods to the train/val/test data in turn.

        Returns:
            If batch size is not set, returns a list of lists of tensors with the inner lists containing [static,
                temporal, labels] for train/val/test respectively.
            If batch size is set returns a list of dataloaders that load from a TensorDataset(temporal, labels) if no
                static data is set, else from StaticTemporalTensorDataset.
        """
        assert all(
            [x is not None for x in self.indices]
        ), "No split indices found, check this has been fitted."

        # Transform the pipelines to each split
        outputs = [
            self._apply_pipeline_to_indices(dataset, indices)
            for indices in self.indices
        ]

        # Convert to dataloaders if specified
        if self.batch_size is not None:
            outputs = [self._to_dataloader(tensors) for tensors in outputs]

        return outputs



import numpy as np


class StaticTemporalTensorDataset:
    """ Outputs a tuple of (static, temporal) and labels at the specified indices. """

    def __init__(self, static_data, temporal_data, labels):
        super(StaticTemporalTensorDataset, self).__init__()
        assert len(static_data) == len(temporal_data) == len(labels)
        self.static_data = static_data
        self.temporal_data = temporal_data
        self.labels = labels

    def __getitem__(self, index):
        return (self.static_data[index], self.temporal_data[index]), self.labels[index]

    def __len__(self):
        return len(self.labels)


def get_num_params(model):
    """Returns the number of trainable parameters in a pytorch model.

    Arguments:
        model (nn.Module): PyTorch model.

    Returns:
        An integer denoting the number of trainable parameters in the model.
    """
    model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    params = sum([np.prod(p.size()) for p in model_parameters])
    return params


def tune_number_of_parameters(model_builder, desired_num_params):
    """Tunes a model parameter to make the total model params as close to desired_num_params as possible.

    Given a function that accepts an integer argument and outputs an initialised nn.Module, increments the argument
    until the number of parameters in the output model is as close to `desired_num_params` as possible.

    Args:
        model_builder (function): A function that accepts an integer argument and outputs a model.
        desired_num_params (int): The number of parameters wanted in the output model.

    Returns:
        An initialised model with number of params close to `desired_num_params`.
    """
    min_params = get_num_params(model_builder(1))
    if min_params > desired_num_params:
        raise ValueError("Minimum number of model params ({}) is more than desired ({})".format(min_params, desired_num_params))

    num_params = 0
    hidden_dim = 1
    while num_params < desired_num_params:
        model = model_builder(hidden_dim)
        num_params = get_num_params(model)
        hidden_dim += 1

    # Revert back one if it is closer
    if hidden_dim - 2 > 0:
        prev_build = model_builder(hidden_dim - 2)
        if desired_num_params - get_num_params(prev_build) < desired_num_params - num_params:
            model = prev_build

    return model



import numpy as np
from sklearn.model_selection import train_test_split


def train_val_test_split(
    tensors,
    val_frac=0.15,
    test_frac=0.15,
    stratify_idx=None,
    shuffle=True,
    seed=None,
    return_indices=False,
):
    """Train test split method for an arbitrary number of tensors.

    Given a list of tensor in the variable `tensors`, splits each randomly into (train_frac, val_frac, test_frac)
    proportions.

    Arguments:
        tensors (list): A list of torch tensors.
        val_frac (float): The fraction to use as validation data.
        test_frac (float): The fraction to use as test data.
        stratify_idx (int): The index of the `tensors` variable to use as stratification labels.
        shuffle (bool): Set True to shuffle first.
        return_indices (bool): Set True to return a 3-tuple of index locations (train_indices, val_indices,
            test_indices) rather than the split tensors themselves.
        seed (int): Random seed.

    Returns:
        A tuple containing three lists corresponding to the train/val/test split of `tensors` or the absolute index
            locations.
    """
    # Set seed
    if seed is not None:
        np.random.seed(seed)

    # Check all tensors have the same length
    num_samples = tensors[0].size(0)
    assert [t.size(0) == num_samples for t in tensors]

    # Add an indices tensor to get absolute index location
    tensors.append(np.arange(len(tensors[0])))

    # Stratification labels
    stratification_labels = None
    if stratify_idx is not None:
        stratification_labels = (
            tensors[stratify_idx] if tensors[stratify_idx].dim() <= 2 else None
        )

    # Get a train+val/test split followed by a train/val split.
    train_val_data, test_data = tensor_train_test_split(
        tensors, test_frac, stratify=stratification_labels, shuffle=shuffle
    )

    # Split out train and val
    if stratify_idx is not None:
        stratification_labels = train_val_data[stratify_idx]
    new_test_frac = val_frac / (1 - test_frac)
    train_data, val_data = tensor_train_test_split(
        train_val_data, new_test_frac, stratify=stratification_labels, shuffle=shuffle
    )

    # Return either of the indices or the full tensors
    tensors = [train_data, val_data, test_data]
    if return_indices:
        output = [x[-1] for x in tensors]
    else:
        output = [x[:-1] for x in tensors]

    return output


def tensor_train_test_split(tensors, test_frac, stratify=None, shuffle=True):
    """Splits a list of tensors into two parts according to the test_frac.

    Arguments:
        tensors (list): A list of tensors.
        test_frac (float): The fraction the test set.
        stratify (tensor): Stratification labels.
        shuffle (bool): Set True to shuffle first.

    Returns:
        Two lists, the first list contains the training tensors, the second the test tensors.
    """
    split_tensors = train_test_split(
        *tensors, stratify=stratify, shuffle=shuffle, test_size=test_frac
    )
    return split_tensors[0::2], split_tensors[1::2]



import torch
from sklearn.base import TransformerMixin
from torch.nn.utils.rnn import pad_sequence


class PadRaggedTensors(TransformerMixin):
    """Converts a list of unequal length tensors (or arrays) to a stacked tensor.

    This is done by extending tensors to the same length as the tensor in the list with maximal length.

    Arguments:
        fill_value (float): Value to fill if an array is extended.

    Returns:
        Two tensors, the first is of shape (len(tensor_list), max length from tensor_list, ...) and is the now stacked
            tensor, the second is the lengths of the tensors.
    """

    def __init__(self, fill_value=float("nan"), max_seq_len=None):
        self.fill_value = fill_value
        self.max_seq_len = max_seq_len

    def fit(self, data, labels=None):
        return self

    def transform(self, data):
        """Pads a list of tensors to max length.

        Args:
            data (list):

        Returns:

        """
        padded_tensor, _ = _pad_ragged_tensors(data, self.fill_value, self.max_seq_len)
        return padded_tensor


def _pad_ragged_tensors(tensor_list, fill_value=float("nan"), max_seq_len=None):
    if not isinstance(tensor_list[0], torch.Tensor):
        tensor_list = [torch.tensor(t) for t in tensor_list]

    # Reduce size
    if max_seq_len is not None:
        tensor_list = [x[:max_seq_len] for x in tensor_list]

    # Pad with a value that doesnt exist in the dataframe
    lengths = [len(x) for x in tensor_list]
    padded_tensor = pad_sequence(
        tensor_list, batch_first=True, padding_value=fill_value
    )

    return padded_tensor, lengths



import pytest
import torch
from torch import nn, optim

from tsbatteries import models
from tsbatteries.models import utils
from tsbatteries.tests.helpers import make_classification_problem

MODELS = {"rnn": models.RNN, "retain": models.RETAIN, "ncde": models.NeuralCDE}


def training_loop(model, data, labels, n_epochs=5):
    labels = labels.to(torch.float)
    optimizer = optim.Adam(model.parameters(), lr=0.1)
    criterion = nn.BCEWithLogitsLoss()
    for _ in range(n_epochs):
        optimizer.zero_grad()
        preds = model(data)
        loss = criterion(preds, labels)
        loss.backward()
        optimizer.step()
    # Evaluate on train
    preds = torch.sigmoid(model(data))
    acc = ((preds - labels) < 0.5).sum() / len(preds)
    return preds, acc


@pytest.mark.parametrize(
    "model_name",
    ["rnn", "retain"],
)
def test_rnn_models(model_name):
    # Test full accuracy on an easy classification problem
    # Note that this additionally tests for standardised inputs
    # Load a basic classification problem
    train_data, _ = make_classification_problem()
    data, labels = train_data
    input_dim, output_dim = data.size(2), labels.size(1)

    # Train the retain model
    model = MODELS[model_name](
        input_dim=input_dim,
        hidden_dim=30,
        output_dim=output_dim,
        return_sequences=False,
    )
    _, acc = training_loop(model, data, labels, n_epochs=50)
    assert 0 <= acc <= 1


def test_tune_number_of_parameters():
    # Load a basic classification problem
    train_data, _ = make_classification_problem(static_dim=None)
    data, labels = train_data
    input_dim, output_dim = data.size(2), labels.size(1)

    # Create a model builder
    def model_builder(x):
        return MODELS["rnn"](
            input_dim=input_dim,
            hidden_dim=x,
            output_dim=output_dim,
            return_sequences=False,
        )

    # Check works
    for num_params in [1000, 50000, 100000]:
        model = utils.tune_number_of_parameters(model_builder, num_params)
        assert 0.9 * num_params < utils.get_num_params(model) < 1.1 * num_params
